# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""vimiv init: tests some imports and imports main."""

import sys

__license__ = "MIT"
__version__ = "0.9.1"
__author__ = __maintainer__ = "Christian Karl"
__email__ = "karlch@protonmail.com"

try:
    from gi import require_version
    require_version("Gtk", "3.0")
    try:
        require_version("GExiv2", "0.10")
    # No EXIF support
    except ValueError:
        pass
    from gi.repository import GLib, Gtk, Gdk, GdkPixbuf, Pango

except ImportError as import_error:
    message = import_error.msg + "\n" + "Are all dependencies installed?"
    print(message)
    sys.exit(1)



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Information pop-up for vimiv.

Shows the current version, the icon, gives a link to the GitHub site and the
possibility to display the licence.
"""

from gi.repository import GLib, Gtk


class Information():
    """Class containing information on vimiv."""

    def get_version(self):
        """Return current version.

        Return:
            String of the current version.
        """
        import vimiv
        return "vimiv version %s" % vimiv.__version__

    def show_version_info(self, running_tests=False):
        """Show information about current version in a Gtk pop-up.

        Args:
            running_tests: If True running from testsuite. Do not show pop-up.
        """
        # Popup with buttons
        popup = Gtk.Dialog(title="vimiv - Info", transient_for=Gtk.Window())
        licence_button = popup.add_button(" Licence", Gtk.ResponseType.HELP)
        popup.add_button("Close", Gtk.ResponseType.CLOSE)
        # Different widgets
        licence_button.connect("clicked", self.show_licence)
        licence_icon = Gtk.Image.new_from_icon_name(
            "text-x-generic-symbolic", 0)
        licence_button.set_image(licence_icon)
        licence_button.set_always_show_image(True)

        version_label = Gtk.Label()
        version_label.set_markup(
            '<span size="xx-large">' + self.get_version() + "</span>")

        info_label = Gtk.Label()
        info_label.set_text("vimiv - an image viewer with vim-like keybindings")
        info_label.set_hexpand(True)

        website_label = Gtk.Label()
        website_label.set_halign(Gtk.Align.END)
        website_label.set_markup('<a href="http://karlch.github.io/vimiv"'
                                 'title="vimiv website">Website</a>')
        github_label = Gtk.Label()
        github_label.set_halign(Gtk.Align.START)
        github_label.set_markup('<a href="https://www.github.com/karlch/vimiv"'
                                'title="vimiv on GitHub">GitHub</a>')

        icon_theme = Gtk.IconTheme.get_default()
        # Icon not available in theme when icon cache was not updated
        try:
            vimiv_pixbuf = icon_theme.load_icon("vimiv", 128, 0)
            vimiv_image = Gtk.Image.new_from_pixbuf(vimiv_pixbuf)
            load_image = True
        except GLib.Error:
            load_image = False

        # Layout
        box = popup.get_child()
        box.set_border_width(12)
        grid = Gtk.Grid()
        grid.set_column_spacing(12)
        grid.set_row_spacing(12)
        box.pack_start(grid, False, False, 12)
        grid.attach(version_label, 0, 0, 2, 1)
        if load_image:
            grid.attach(vimiv_image, 0, 1, 2, 1)
        grid.attach(info_label, 0, 2, 2, 1)
        grid.attach(website_label, 0, 3, 1, 1)
        grid.attach(github_label, 1, 3, 1, 1)
        popup.show_all()
        if not running_tests:
            popup.run()
            popup.destroy()

    def show_licence(self, button, running_tests=False):
        """Show the licence in a new pop-up window.

        Args:
            button: The Gtk.Button() that called this function.
            running_tests: If True running from testsuite. Do not show pop-up.
        """
        popup = Gtk.Dialog(title="vimiv - Licence", transient_for=Gtk.Window())
        popup.add_button("Close", Gtk.ResponseType.CLOSE)

        label = Gtk.Label()
        title = '<span size="large"><b>MIT Licence</b></span>\n\n'

        # Get license text from file
        with open("/usr/share/licenses/vimiv/LICENSE") as license_file:
            license_text = license_file.read()
        label_markup = title + license_text
        label.set_markup(label_markup)

        box = popup.get_child()
        box.pack_start(label, False, False, 0)
        box.set_border_width(12)
        popup.show_all()
        if not running_tests:
            popup.run()
            popup.destroy()



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Deal with the handling of tags for vimiv."""

import os

from gi.repository import GLib, Gtk
from vimiv.helpers import error_message, read_file


class TagHandler(object):
    """Handle tags.

    Attributes:
        directory: Directory in which tags are stored.
        last: Last tag that was loaded.

        _app: The main vimiv application to interact with.
    """

    def __init__(self, app):
        """Create the necessary objects.

        Args:
            app: The main vimiv application to interact with.
        """
        self._app = app
        # Create tags directory
        self.directory = os.path.join(GLib.get_user_data_dir(), "vimiv", "Tags")
        os.makedirs(self.directory, exist_ok=True)
        # If there are tags in the old location, move them to the new one
        # TODO remove this in a new version, assigned for 0.10
        old_directory = os.path.expanduser("~/.vimiv/Tags")
        if os.path.isdir(old_directory):
            old_tags = os.listdir(old_directory)
            for f in old_tags:
                old_path = os.path.join(old_directory, f)
                new_path = os.path.join(self.directory, f)
                os.rename(old_path, new_path)
            message = "Moved tags from the old directory %s " \
                "to the new location %s." % (old_directory, self.directory)
            error_message(message, running_tests=self._app.running_tests)
            os.rmdir(old_directory)
        self.last = ""

    def write(self, imagelist, tagname):
        """Add a list of images to a tag.

        Args:
            imagelist: List of images to write to the tag.
            tagname: Name of tag to operate on.
        """
        # Backup required for tests
        imagelist = imagelist if imagelist else self._app["mark"].marked
        tagfile_name = os.path.join(self.directory, tagname)
        tagged_images = self._read(tagname)
        with open(tagfile_name, "a") as tagfile:
            for image in imagelist:
                if image not in tagged_images:
                    tagfile.write(image + "\n")

    def remove(self, tagname):
        """Remove a tag showing an error if the tag doesn't exist.

        Args:
            tagname: Name of tag to operate on.
        """
        tagfile_name = os.path.join(self.directory, tagname)
        if os.path.isfile(tagfile_name):
            os.remove(tagfile_name)
        else:
            err = "Tagfile '%s' does not exist" % (tagname)
            self._app["statusbar"].message(err, "error")

    def load(self, tagname):
        """Load all images in a tag as current filelist.

        Args:
            tagname: Name of tag to operate on.
        """
        os.chdir(self.directory)
        # Read file and get all tagged images as list
        tagged_images = self._read(tagname)
        # Populate filelist
        self._app.populate(tagged_images, expand_single=False)
        if self._app.get_paths():
            self._app["main_window"].show()
            self._app["library"].set_hexpand(False)
            self._app["image"].load()
            # Focus in library if it is open
            if self._app["library"].grid.is_visible():
                self._app["library"].reload(self.directory)
                tag_pos = self._app["library"].files.index(tagname)
                self._app["library"].set_cursor(Gtk.TreePath(tag_pos),
                                                None, False)
            # Remember last tag selected
            self.last = tagname
        else:
            message = "Tagfile '%s' has no valid images" % (tagname)
            self._app["statusbar"].message(message, "error")
            self.last = ""

    def _read(self, tagname):
        """Read a tag and return the images in it.

        Args:
            tagname: Name of tag to operate on.

        Return:
            List of images in the tag called tagname.
        """
        tagfile_name = os.path.join(self.directory, tagname)
        tagged_images = read_file(tagfile_name)
        return tagged_images



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""The commandline completion for vimiv."""

import os
import re
from itertools import takewhile
from string import digits

from gi.repository import GLib, Gtk
from vimiv.fileactions import is_image
from vimiv.helpers import listdir_wrapper, read_info_from_man
from vimiv.settings import settings


class Completion(Gtk.TreeView):
    """Completion for vimiv's commandline.

    Attributes:
        info: ScrolledWindow containing the completion information.

        _app: The main vimiv application to interact with.
        _liststores: Dictionary containing the liststore models for different
            completion types.
        _prefixed_digits: User prefixed digits for internal command.
        _tab_position: Position when tabbing through completion elements.
        _tab_presses: Amount of tab presses.
    """

    def __init__(self, app):
        """Create the necessary objects and settings.

        Args:
            app: The main vimiv application to interact with.
        """
        super(Completion, self).__init__()
        self._app = app
        # Create liststore dictionary with the filter used.
        self._liststores = {"internal": [Gtk.ListStore(str, str)],
                            "external": [Gtk.ListStore(str, str)],
                            "path": [Gtk.ListStore(str, str)],
                            "tag": [Gtk.ListStore(str, str)],
                            "search": [Gtk.ListStore(str, str)],
                            "trash": [Gtk.ListStore(str, str)],
                            "setting": [Gtk.ListStore(str, str)]}
        for liststore in self._liststores.values():
            comp_filter = liststore[0].filter_new()
            comp_filter.set_visible_func(self._completion_filter)
            liststore.append(comp_filter)
        # Use the commandline entry here and connect to the refilter method
        self._app["commandline"].connect("changed", self._refilter)
        # Edit treeview
        self.set_enable_search(False)
        self.set_headers_visible(False)
        self.set_activate_on_single_click(True)
        padding = settings["commandline_padding"].get_value()
        renderer = Gtk.CellRendererText()
        renderer.set_padding(padding, 0)
        command_column = Gtk.TreeViewColumn("Command", renderer, markup=0)
        command_column.set_expand(True)
        self.append_column(command_column)
        info_column = Gtk.TreeViewColumn("Info", renderer, markup=1)
        self.append_column(info_column)
        self.connect("row-activated", self._activate)
        # Scrolled window for the completion info
        self.info = Gtk.ScrolledWindow()
        self.info.set_size_request(
            10, settings["completion_height"].get_value())
        self.info.add(self)
        # Defaults
        self._tab_position = 0
        self._tab_presses = 0
        self._prefixed_digits = ""

    def generate_commandlist(self):
        """Generate a list of internal vimiv commands and store it."""
        commands = [cmd.name for cmd in self._app["commandline"].commands
                    if not cmd.is_hidden]
        aliases = list(self._app["commandline"].commands.aliases.keys())
        all_commands = sorted(commands + aliases)
        infodict = read_info_from_man()
        for command in all_commands:
            if command in infodict:
                info = infodict[command]
            elif command in aliases:
                info = "Alias to %s" % (
                    self._app["commandline"].commands.aliases[command])
            else:
                info = ""
            info = "<i>" + GLib.markup_escape_text(info) + "</i>"
            self._liststores["internal"][0].append([command, info])

    def show(self):
        """Show the completion information."""
        # Hacky way to not show the last selected item
        self.set_model(Gtk.ListStore(str, str))
        self._refilter(self._app["commandline"])
        self.info.show()

    def hide(self):
        """Hide the completion information."""
        self.info.hide()

    def activate_tab_completion(self):
        """Select tab completion when commandline was activated.

        Return:
            Boolean: True if completion was selected, False else.
        """
        if self._tab_presses:
            self._activate(None, None, None)
            self.reset()
            return True
        return False

    def _activate(self, treeview, path, column):
        """Enter the completion text of the treeview into the commandline.

        Args:
            treeview: TreeView that was activated.
            path: Activated TreePath.
            column: Activated TreeViewColumn.
        """
        if treeview:
            count = path.get_indices()[0]
            self._app["commandline"].grab_focus()
        else:
            count = self._tab_position
        comp_type = self._get_comp_type()
        row = self._liststores[comp_type][1][count]
        self._app["commandline"].set_text(":" + self._prefixed_digits + row[0])
        self._app["commandline"].set_position(-1)

    def reset(self):
        """Reset all internal counts."""
        self._tab_position = 0
        self._tab_presses = 0

    def complete(self, inverse=False):
        """Run completion.

        Try to enter the best matching completion into the entry. On more than
        one tab start to run through the possible completions.

        Args:
            inverse: If True, tabbing backwards.
        """
        maximum = len(self.get_model())
        # If we have no entries, completing makes no sense
        if not maximum:
            return
        # Try to set best match first
        elif not self._tab_presses:
            best_match = ":" + self._prefixed_digits
            comp_type = self._get_comp_type()
            liststore = self._liststores[comp_type][1]
            first = str(liststore[0][0])
            last = str(liststore[-1][0])
            for i, char in enumerate(first):
                if char == last[i]:
                    best_match += char
                else:
                    break
            if best_match != self._app["commandline"].get_text():
                self._app["commandline"].set_text(best_match)
                self._app["commandline"].set_position(-1)
                return
            # Start at the last element with Shift+Tab
            elif inverse:
                self._tab_position = -1
        # Set position according to direction and move
        elif inverse:
            self._tab_position -= 1
        else:
            self._tab_position += 1
        self._tab_presses += 1
        self._tab_position %= maximum
        self.set_cursor(Gtk.TreePath(self._tab_position), None, False)
        self.scroll_to_cell(Gtk.TreePath(self._tab_position),
                            None, True, 0.5, 0)
        return True  # Deactivate default keybinding (here for Tab)

    def _get_comp_type(self, command=""):
        """Get the current completion type depending on command.

        Args:
            command: The command to check completion type for. If there is none,
                default to getting the text from self._app["commandline"].
        Return:
            The completion type to use.
        """
        if not command:
            command = self._app["commandline"].get_text()
        if command and command[0] != ":":
            return "search"
        command = command.lstrip(":")
        # Nothing entered -> checks are useless
        if command:
            # External commands are prefixed with an !
            if command[0] == "!":
                return "external"
            # Paths are prefixed with a /
            elif command[0] in ["/", ".", "~"]:
                return "path"
            # Tag commands
            elif re.match(r'^(tag_(write|remove|load) )', command):
                return "tag"
            # Undelete files from trash
            elif command.startswith("undelete"):
                return "trash"
            # Settings
            elif command.startswith("set"):
                return "setting"
        return "internal"

    def _complete_path(self, path):
        """Complete paths.

        Args:
            path: (Partial) name of the path to run completion on.
        Return:
            List containing formatted matching paths.
        """
        self._liststores["path"][0].clear()
        # Directory of the path, default to .
        directory = os.path.dirname(path) if os.path.dirname(path) else path
        if not os.path.exists(os.path.expanduser(directory)) \
                or not os.path.isdir(os.path.expanduser(directory)):
            directory = "."
        # /dev causes havoc
        if directory == "/dev":
            return
        # Files in that directory
        files = listdir_wrapper(directory, settings["show_hidden"].get_value())
        # Format them neatly depending on directory and type
        for fil in files:
            fil = os.path.join(directory, fil)
            # Directory
            if os.path.isdir(os.path.expanduser(fil)):
                self._liststores["path"][0].append([fil + "/", ""])
            # Acceptable file
            elif is_image(fil):
                self._liststores["path"][0].append([fil, ""])

    def _complete_tag(self, command):
        """Append the available tag names to an internal tag command.

        Args:
            command: The internal command to complete.
        """
        self._liststores["tag"][0].clear()
        tags = listdir_wrapper(self._app["tags"].directory,
                               settings["show_hidden"].get_value())
        for tag in tags:
            self._liststores["tag"][0].append([command.split()[0] + " " + tag,
                                               ""])

    #  The function currently only hides the commandline but is implemented
    #  for possible future usage.
    def _complete_search(self):
        """Hide the info as it has no use in search."""
        self.info.hide()

    def _complete_external(self, command):
        """If there is more than one word in command, do path completion.

        Args:
            command: The external command to complete.
        """
        self._liststores["external"][0].clear()
        command = command.lstrip("!").lstrip(" ")
        spaces = command.count(" ")
        # If there are spaces, path completion for external commands makes sense
        if spaces:
            args = command.split()
            # Last argument is the path to complete
            if len(args) > spaces:
                path = args[-1]
                # Do not try path completion for arguments
                if path.startswith("-"):
                    self.info.hide()
                    return
                command = " ".join(args[:-1]) + " "
            else:
                path = "./"
            self._complete_path(path)
            # Prepend command to path completion
            # Gtk ListStores are iterable
            # pylint:disable=not-an-iterable
            for row in self._liststores["path"][0]:
                match = "!" + command + re.sub(r'^\./', "", row[0])
                self._liststores["external"][0].append([match, ""])
        else:
            self.info.hide()

    def _complete_trash(self, command):
        """Complete files in trash directory for :undelete.

        Args:
            command: The internal command to complete.
        """
        # Get files in trash directory
        self._liststores["trash"][0].clear()
        trash_directory = \
            self._app["transform"].trash_manager.get_files_directory()
        trash_files = sorted(os.listdir(trash_directory))
        # Add them to completion formatted to 'undelete $FILE'
        for fil in trash_files:
            # Ensure we only complete image files as vimiv is not meant as a
            # general trash tool but provides this for convenience
            abspath = os.path.join(trash_directory, fil)
            if is_image(abspath):
                completion = "undelete %s" % (fil)
                self._liststores["trash"][0].append([completion, ""])

    def _complete_setting(self, command):
        """Complete settings for :set.

        Args:
            command: The internal command to complete.
        """
        self._liststores["setting"][0].clear()
        # Get all settings
        setting_names = [setting.name for setting in settings]
        # Add them to completion formatted to 'set $SETTING'
        for name in setting_names:
            completion = "set %s" % (name)
            typestr = str(type(settings[name].get_value()))
            typestr = typestr.replace("<class '", "").rstrip("'>")
            info = "<i>" + typestr + "</i>"
            self._liststores["setting"][0].append([completion, info])

    def _completion_filter(self, model, treeiter, data):
        """Filter function used in the liststores to filter completions.

        Args:
            model: The liststore model to filter.
            treeiter: The TreeIter representing a row in the model.
            data: User appended data, here the completion mode.
        Return:
            True for a match, False else.
        """
        command = self._app["commandline"].get_text().lstrip(":")
        # Allow number prefixes
        self._prefixed_digits = "".join(takewhile(str.isdigit, command))
        command = command.lstrip(digits)
        return model[treeiter][0].startswith(command)

    def _refilter(self, entry):
        """Refilter the completion list according to the text in entry."""
        # Replace multiple spaces with single space
        entry.set_text(re.sub(r' +', " ", entry.get_text()))
        comp_type = self._get_comp_type()
        command = entry.get_text().lstrip(":")
        if command:
            self.info.show()
        if comp_type == "path":
            self._complete_path(command)
        elif comp_type == "tag":
            self._complete_tag(command)
        elif comp_type == "search":
            self._complete_search()
        elif comp_type == "external":
            self._complete_external(command)
        elif comp_type == "trash":
            self._complete_trash(command)
        elif comp_type == "setting":
            self._complete_setting(command)
        self.set_model(self._liststores[comp_type][1])
        self._liststores[comp_type][1].refilter()
        self.reset()



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Different actions applying directly to files."""

import os
from random import shuffle

from gi.repository import Gdk, GdkPixbuf, Gtk
from vimiv.helpers import listdir_wrapper
from vimiv.settings import settings

# We need the try ... except wrapper here
# pylint: disable=ungrouped-imports
try:
    from gi.repository import GExiv2
    _has_exif = True
except ImportError:
    _has_exif = False


def recursive_search(directory):
    """Search a directory recursively for images.

    Args:
        directory: Directory to search for images.
    Return:
        List of images in directory.
    """
    for root, _, files in os.walk(directory):
        for fil in files:
            yield os.path.join(root, fil)


def populate_single(arg, recursive):
    """Populate a complete filelist if only one path is given.

    Args:
        arg: Single path given.
        recursive: If True search path recursively for images.
    Return:
        Generated list of paths.
    """
    paths = []
    if os.path.isfile(arg):
        # Use parent directory
        directory = os.path.dirname(arg)
        if not directory:  # Default to current directory
            directory = "./"
        paths = listdir_wrapper(directory)
        paths = [os.path.join(directory, path) for path in paths]
        # Set the argument to the beginning of the list
    elif os.path.isdir(arg) and recursive:
        paths = sorted(recursive_search(arg))
    return paths


def populate(args, recursive=False, shuffle_paths=False, expand_single=True):
    """Populate a list of files out given paths.

    Args:
        args: Paths given.
        recursive: If True search path recursively for images.
        shuffle_paths: If True shuffle found paths randomly.
        expand_single: If True, populate a complete filelist with images from
            the same directory as the single argument given.
    Return:
        Found paths, position of first given path.
    """
    paths = []
    index = 0
    # If only one path is passed do special stuff
    first_path = os.path.abspath(args[0]) if args else None
    if len(args) == 1 and expand_single:
        args = populate_single(first_path, recursive)

    # Add everything
    for arg in args:
        path = os.path.abspath(arg)
        if os.path.isfile(path):
            paths.append(path)
        elif os.path.isdir(path) and recursive:
            paths = list(recursive_search(path))
    # Remove unsupported files
    paths = [possible_path for possible_path in paths
             if is_image(possible_path)]
    index = paths.index(first_path) if first_path in paths else 0

    # Shuffle
    if shuffle_paths:
        shuffle(paths)

    return paths, index


def is_image(filename):
    """Check whether a file is an image.

    Args:
        filename: Name of file to check.
    """
    complete_name = os.path.abspath(os.path.expanduser(filename))
    return bool(GdkPixbuf.Pixbuf.get_file_info(complete_name)[0])


def is_animation(filename):
    """Check whether a file is an animated image.

    Args:
        filename: Name of file to check.
    """
    complete_name = os.path.abspath(os.path.expanduser(filename))
    info = GdkPixbuf.Pixbuf.get_file_info(complete_name)[0]
    return "gif" in info.get_extensions()


def is_svg(filename):
    """Check whether a file is a vector graphic.

    Args:
        filename: Name of file to check.
    """
    complete_name = os.path.abspath(os.path.expanduser(filename))
    info = GdkPixbuf.Pixbuf.get_file_info(complete_name)[0]
    return "svg" in info.get_extensions() if info else False


def edit_supported(filename):
    """Check whether a file is editable for vimiv.

    Args:
        filename: Name of file to check.
    """
    complete_name = os.path.abspath(os.path.expanduser(filename))
    info = GdkPixbuf.Pixbuf.get_file_info(complete_name)[0]
    extension = info.get_extensions()[0]
    if extension in ["jpeg", "png", "tiff", "ico", "bmp"]:
        return True
    return False


def format_files(app, string):
    """Format image names in filelist according to a formatstring.

    Numbers files in form of formatstring_000.extension. Replaces exif
    information accordingly.

    Args:
        app: Vimiv application to interact with.
        string: Formatstring to use.
    """
    # Catch problems
    if app["library"].is_focus():
        message = "Format only works on opened image files"
        app["statusbar"].message(message, "info")
        return
    if not app.get_paths():
        app["statusbar"].message("No files in path", "info")
        return

    # Check if exif data is available and needed
    tofind = ("%" in string)
    if tofind:
        if not _has_exif:
            app["statusbar"].message(
                "Install gexiv2 for EXIF support in vimiv", "error")
            return
        for fil in app.get_paths():
            exif = GExiv2.Metadata(fil)
            try:
                exif.get_date_time()
            except KeyError:
                app["statusbar"].message(
                    "No exif data for %s available" % (fil), "error")
                return

    for i, fil in enumerate(app.get_paths()):
        ending = os.path.splitext(fil)[1]
        num = "%03d" % (i + 1)
        # Exif stuff
        if tofind:
            exif = GExiv2.Metadata(fil)
            date = exif.get_date_time()
            outstring = string.replace("%Y", str(date.year))
            outstring = outstring.replace("%m", str(date.month))
            outstring = outstring.replace("%d", str(date.day))
            outstring = outstring.replace("%H", str(date.hour))
            outstring = outstring.replace("%M", str(date.minute))
            outstring = outstring.replace("%S", str(date.second))
        else:
            outstring = string
        # Ending
        outstring += num + ending
        os.rename(fil, outstring)

    app.emit("paths-changed", format_files)


class ClipboardHandler(object):
    """Deals with copying to the system clipboard."""

    def __init__(self, app):
        """Receive and set main vimiv application.

        Args:
            _app: The main vimiv class to interact with.
        """
        self._app = app

    def copy_name(self, abspath=False):
        """Copy image name to clipboard.

        Args:
            abspath: Use absolute path or only the basename.
        """
        # Get name to copy
        name = self._app.get_pos(True)
        if abspath:
            name = os.path.abspath(name)
        else:
            name = os.path.basename(name)
        # Set clipboard
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_PRIMARY) \
            if settings["copy_to_primary"].get_value() \
            else Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        # Text to clipboard
        clipboard.set_text(name, -1)
        # Info message
        message = "Copied <b>" + name + "</b> to %s" % \
            ("primary" if settings["copy_to_primary"].get_value()
             else "clipboard")
        self._app["statusbar"].message(message, "info")



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Contains custom exceptions used by vimiv."""


class NoSearchResultsError(Exception):
    """Raised when a search result is accessed although there are no results."""


class StringConversionError(ValueError):
    """Raised when a setting or argument could not be converted to its type."""


class SettingNotFoundError(Exception):
    """Raised when a setting does not exist."""


class NotABoolean(Exception):
    """Raised when a setting is not a boolean."""


class NotANumber(Exception):
    """Raised when a setting is not a number."""


class AliasError(Exception):
    """Raised when there are problems when adding an alias."""


class TrashUndeleteError(Exception):
    """Raised when there were problems calling :undelete."""


class NotTransformable(Exception):
    """Raised when an image is not transformable for transform.py."""


class ArgumentAmountError(Exception):
    """Raised if the amount of arguments is not compatible with a command."""



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Gtk.ScrolledWindow class which is usually the main window of vimiv.

The ScrolledWindow can either include a Gtk.Image in IMAGE mode or a
Gtk.IconView in THUMBNAIL mode.
"""

import os

from gi.repository import Gtk
from vimiv.helpers import listdir_wrapper
from vimiv.image import Image
from vimiv.thumbnail import Thumbnail


class MainWindow(Gtk.ScrolledWindow):
    """Main window of vimiv containing either an Image or an IconView.

    Attributes:
        image: Vimiv Image class which may be displayed.
        thumbnail: Vimiv Thumbnail class which may be displayed.

        _app: The main vimiv class to interact with.
    """

    def __init__(self, app):
        """Initialize image and thumbnail attributes and configure self."""
        super(MainWindow, self).__init__()
        self._app = app
        self.image = Image(app)
        self.thumbnail = Thumbnail(app)

        self.set_hexpand(True)
        self.set_vexpand(True)
        # Image is default
        self.add(self.image)
        self.connect("key_press_event", self._app["eventhandler"].on_key_press,
                     "IMAGE")

        # Connect signals
        self._app.connect("widget-layout-changed", self._on_widgets_changed)
        self._app.connect("paths-changed", self._on_paths_changed)

    def switch_to_child(self, new_child):
        """Switch the widget displayed in the main window.

        Args:
            new_child: The child to switch to.
        """
        self.remove(self.get_child())
        self.add(new_child)
        self.show_all()

    def center_window(self):
        """Center the widget in the current window."""
        h_adj = self.get_hadjustment()
        size = self.get_allocation()
        h_middle = (h_adj.get_upper() - h_adj.get_lower() - size.width) / 2
        h_adj.set_value(h_middle)
        v_adj = self.get_vadjustment()
        v_middle = (v_adj.get_upper() - v_adj.get_lower() - size.height) / 2
        v_adj.set_value(v_middle)
        self.set_hadjustment(h_adj)
        self.set_vadjustment(v_adj)

    def scroll(self, direction):
        """Scroll the correct object.

        Args:
            direction: Scroll direction to emit.
        """
        if direction not in "hjklHJKL":
            self._app["statusbar"].message(
                "Invalid scroll direction " + direction, "error")
        elif self.thumbnail.toggled:
            self.thumbnail.move_direction(direction)
        else:
            self._scroll(direction)
        return True  # Deactivates default bindings (here for Arrows)

    def _scroll(self, direction):
        """Scroll the widget.

        Args:
            direction: Direction to scroll in.
        """
        steps = self._app["eventhandler"].num_receive()
        scale = self.image.get_scroll_scale()
        h_adj = self.get_hadjustment()
        size = self.get_allocation()
        h_size = h_adj.get_upper() - h_adj.get_lower() - size.width
        h_step = h_size / scale * steps
        v_adj = self.get_vadjustment()
        v_size = v_adj.get_upper() - v_adj.get_lower() - size.height
        v_step = v_size / scale * steps
        # To the ends
        if direction == "H":
            h_adj.set_value(0)
        elif direction == "J":
            v_adj.set_value(v_size)
        elif direction == "K":
            v_adj.set_value(0)
        elif direction == "L":
            h_adj.set_value(h_size)
        # By step
        elif direction == "h":
            h_adj.set_value(h_adj.get_value() - h_step)
        elif direction == "j":
            v_adj.set_value(v_adj.get_value() + v_step)
        elif direction == "k":
            v_adj.set_value(v_adj.get_value() - v_step)
        elif direction == "l":
            h_adj.set_value(h_adj.get_value() + h_step)
        self.set_hadjustment(h_adj)
        self.set_vadjustment(v_adj)

    def _on_widgets_changed(self, app, widget):
        """Recalculate thumbnails or rezoom image when the layout changed."""
        if self.thumbnail.toggled:
            self.thumbnail.calculate_columns()
        elif self._app.get_paths() and self.image.fit_image != "user":
            self.image.zoom_to(0, self.image.fit_image)

    def _on_paths_changed(self, app, transform):
        """Reload paths image and/or thumbnail when paths have changed."""
        if self._app.get_paths():
            # Get all files in directory again
            focused_path = self._app.get_pos(True)
            decremented_index = max(0, self._app.get_pos() - 1)
            directory = os.path.dirname(focused_path)
            files = [os.path.join(directory, fil)
                     for fil in listdir_wrapper(directory)]
            self._app.populate(files)
            # Reload thumbnail
            if self.thumbnail.toggled:
                self.thumbnail.on_paths_changed()
            # Refocus the path
            if focused_path in self._app.get_paths():
                index = self._app.get_paths().index(focused_path)
            # Stay as close as possible
            else:
                index = min(decremented_index, len(self._app.get_paths()) - 1)
            if self.thumbnail.toggled:
                self.thumbnail.move_to_pos(index)
            else:
                self._app["eventhandler"].set_num_str(index + 1)
                self.image.move_pos()
            self._app["statusbar"].update_info()
        # We need to check again as populate was called
        if not self._app.get_paths():
            self.hide()



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Deals with transformations like rotate and flip and deleting files."""

import os
from threading import Thread

from gi.repository import GObject
from vimiv import imageactions
from vimiv.exceptions import (NotTransformable, TrashUndeleteError,
                              StringConversionError)
from vimiv.fileactions import edit_supported
from vimiv.helpers import get_int
from vimiv.settings import settings
from vimiv.trash_manager import TrashManager


class Transform(GObject.Object):
    """Deals with transformations like rotate/flip and deleting files.

    Attributes:
        threads_running: If True, a thread is running to apply to files.
        trash_manager: Class to handle a shared trash directory.

        _app: The main vimiv application to interact with.
        _changes: Dictionary for rotate and flip.
            Key: Filename; Item: [Int, Bool, Bool]

    Signals:
        changed: Emitted when an image was transformed so Image can update.
        applied-to-file: Emitted when the file was successfully transformed.
    """

    def __init__(self, app):
        super(Transform, self).__init__()
        self._app = app
        self._changes = {}
        self.trash_manager = TrashManager()
        self.threads_running = False

    def delete(self):
        """Delete all marked images or the current one."""
        # Get all images
        images = self.get_images("Deleted")
        self._app["mark"].marked = []
        # Delete all images remembering possible errors
        message = ""
        for im in images:
            if not os.path.exists(im):
                message += "Image %s does not exist." % (im)
            elif os.path.isdir(im):
                message += "Deleting directory %s is not supported." % (im)
            else:
                self.trash_manager.delete(im)
        if message:
            self._app["statusbar"].message(message, "error")

        self._app.emit("paths-changed", self)

    def undelete(self, basename):
        """Undelete an image in the trash.

        Args:
            basename: The basename of the image in the trash directory.
        """
        try:
            self.trash_manager.undelete(basename)
            self._app.emit("paths-changed", self)
        except TrashUndeleteError as e:
            message = "Could not restore %s, %s" % (basename, str(e))
            self._app["statusbar"].message(message, "error")

    def write(self, quit_app=False):
        """Write transformed/manipulated image(s) to disk.

        Args:
            quit: If True, quit the application. Activated by :wq.
        """
        self._app["statusbar"].message("Saving...", "info")
        # Manipulations include transformations implicitly
        if self._app["manipulate"].is_visible():
            self._app["manipulate"].finish(True)
            self._changes.clear()
        # Only apply any transformations
        else:
            self._thread_for_apply()
        # Quit or inform
        if quit_app:
            self._app.quit_wrapper()
        else:
            self._app["statusbar"].message("Changes written to disk", "info")

    def get_images(self, info):
        """Return the images which should be manipulated.

        Either the currently focused image or all marked images.

        Args:
            info: Info to display when acting on marked images.
        """
        # Add all marked images
        if self._app["mark"].marked:
            images = self._app["mark"].marked
            if len(images) == 1:
                message = "%s %d marked image" % (info, len(images))
            else:
                message = "%s %d marked images" % (info, len(images))
            self._app["statusbar"].message(message, "info")
        # Add the image shown
        else:
            images = [os.path.abspath(self._app.get_pos(True))]
        return images

    def rotate(self, cwise):
        """Rotate the displayed image and call thread to rotate files.

        Args:
            cwise: Rotate image 90 * cwise degrees.
        """
        try:
            self._is_transformable()
            cwise = get_int(cwise, allow_sign=True)
        except NotTransformable as e:
            self._app["statusbar"].message(str(e) + " rotate", "error")
            return
        except StringConversionError as e:
            self._app["statusbar"].message(str(e), "error")
            return
        images = self.get_images("Rotated")
        cwise = cwise % 4
        # Update properties
        for fil in images:
            if fil in self._changes:
                self._changes[fil][0] = \
                    (self._changes[fil][0] + cwise) % 4
            else:
                self._changes[fil] = [cwise, 0, 0]
        # Rotate the image shown
        if self._app.get_path() in images:
            self.emit("changed", "rotate", cwise)
        # Reload thumbnails of rotated images immediately
        if self._app["thumbnail"].toggled:
            self.apply()

    def apply(self):
        """Start thread for rotate and flip."""
        # TODO improve this, it is currently not possible to find out what is
        # being changed and what should still be done
        if self.threads_running:
            return
        if settings["autosave_images"].get_value():
            t = Thread(target=self._thread_for_apply)
            t.start()
        else:
            self._changes.clear()

    def _thread_for_apply(self):
        """Rotate and flip image file in an extra thread."""
        self.threads_running = True
        to_remove = list(self._changes.keys())
        for f in self._changes:
            if self._changes[f][0]:
                imageactions.rotate_file(f, self._changes[f][0])
            if self._changes[f][1]:
                imageactions.flip_file(f, True)
            if self._changes[f][2]:
                imageactions.flip_file(f, False)
        for key in to_remove:
            del self._changes[key]
        self.emit("applied-to-file", to_remove)
        self.threads_running = False

    def _is_transformable(self):
        """Check if the current image is transformable."""
        if not self._app.get_paths():
            raise NotTransformable("No image to")
        elif not edit_supported(self._app.get_path()):
            raise NotTransformable("Filetype not supported for")
        # Some operations only make sense if we are allowed to save to file
        elif not settings["autosave_images"].get_value():
            message = ""
            if self._app["thumbnail"].toggled:
                message = 'When operating in thumbnail mode ' \
                          '"autosave_images" must be enabled for'
            elif self._app["mark"].marked:
                message = 'When images are marked ' \
                          '"autosave_images" must be enabled for'
            if message:
                raise NotTransformable(message)

    def flip(self, horizontal):
        """Flip the displayed image and call thread to flip files.

        Args:
            horizontal: If True, flip horizontally. Else vertically.
        """
        try:
            self._is_transformable()
            horizontal = get_int(horizontal)
        except NotTransformable as e:
            self._app["statusbar"].message(str(e) + " flip", "error")
            return
        except StringConversionError as e:
            self._app["statusbar"].message(str(e), "error")
            return
        images = self.get_images("Flipped")
        # Apply changes
        for fil in images:
            if fil not in self._changes:
                self._changes[fil] = [0, 0, 0]
            if horizontal:
                self._changes[fil][1] = \
                    (self._changes[fil][1] + 1) % 2
            else:
                self._changes[fil][2] = \
                    (self._changes[fil][2] + 1) % 2
        # Flip the image shown
        if self._app.get_path() in images:
            self.emit("changed", "flip", horizontal)
        # Reload thumbnails of flipped images immediately
        if self._app["thumbnail"].toggled:
            self.apply()

    def rotate_auto(self):
        """Autorotate all pictures in the current pathlist."""
        autorotate = imageactions.Autorotate(self._app.get_paths())
        self.threads_running = True
        autorotate.connect("completed", self._on_autorotate_completed)
        autorotate.run()

    def _on_autorotate_completed(self, autorotate, amount):
        message = "Completed autorotate, %d files rotated" % (amount)
        self.threads_running = False
        self._app["statusbar"].message(message, "info")


GObject.signal_new("changed", Transform, GObject.SIGNAL_RUN_LAST, None,
                   (GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT))
GObject.signal_new("applied-to-file", Transform, GObject.SIGNAL_RUN_LAST, None,
                   (GObject.TYPE_PYOBJECT,))



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Provides classes to store and load thumbnails from a shared cache.

The ThumbnailStore transparently creates and loads thumbnails according to the
freedesktop.org thumbnail management standard.
The ThumbnailManager provides a asynchronous mechanism to load thumbnails from
the store.

If possible, you should avoid using the store directly but use the manager
instead.
"""

import collections
import hashlib
import os
import tempfile
from multiprocessing.pool import ThreadPool as Pool

from gi._error import GError
from gi.repository import GdkPixbuf, GLib, Gtk
from gi.repository.GdkPixbuf import Pixbuf

ThumbTuple = collections.namedtuple('ThumbTuple', ['original', 'thumbnail'])


class ThumbnailManager:
    """Provides an asynchronous mechanism to load thumbnails.

    Attributes:
        thumbnail_store: ThumbnailStore class with the loading mechanism.
        large: The thumbnail managing standard specifies two thumbnail sizes.
            256x256 (large) and 128x128 (normal)
        default_icon: Default icon if thumbnails are not yet loaded.
        error_icon: The path to the icon which is used, when thumbnail creation
                    fails.
    """

    _cpu_count = os.cpu_count()
    if _cpu_count is None:
        _cpu_count = 1
    elif _cpu_count > 1:
        _cpu_count -= 1

    _thread_pool = Pool(_cpu_count)
    _cache = {}

    def __init__(self, large=True):
        """Construct a new ThumbnailManager.

        Args:
            large: Size of thumbnails that are created. If true 256x256 else
                   128x128.
        """
        super(ThumbnailManager, self).__init__()
        self.thumbnail_store = ThumbnailStore(large=large)

        # Default icon if thumbnail creation fails
        icon_theme = Gtk.IconTheme.get_default()
        self.error_icon = icon_theme.lookup_icon("dialog-error", 256,
                                                 0).get_filename()
        self.default_icon = icon_theme.lookup_icon("image-x-generic", 256,
                                                   0).get_filename()

    def _do_get_thumbnail_at_scale(self, source_file, size, callback, index,
                                   ignore_cache=False):
        if not ignore_cache and source_file in self._cache:
            pixbuf = self._cache[source_file]
        else:
            thumbnail_path = self.thumbnail_store.get_thumbnail(source_file,
                                                                ignore_cache)
            if thumbnail_path is None:
                thumbnail_path = self.error_icon
            pixbuf = Pixbuf.new_from_file(thumbnail_path)
            self._cache[source_file] = pixbuf

        if pixbuf.get_height() != size and pixbuf.get_width != size:
            pixbuf = self.scale_pixbuf(pixbuf, size)

        return callback, pixbuf, index

    @staticmethod
    def scale_pixbuf(pixbuf, size):
        """Scale the pixbuf to the given size keeping the aspect ratio.

        Either the width or the height of the returned pixbuf is `size` large,
        depending on the aspect ratio.

        Args:
            pixbuf: The pixbuf to scale
            size: The size of the new width or height

        Return:
            The scaled pixbuf.
        """
        width = size
        height = size
        ratio = pixbuf.get_width() / pixbuf.get_height()
        if ratio > 1:
            height /= ratio
        else:
            width *= ratio

        pixbuf = pixbuf.scale_simple(width, height,
                                     GdkPixbuf.InterpType.BILINEAR)
        return pixbuf

    @staticmethod
    def _do_callback(result):
        GLib.idle_add(*result)

    def get_thumbnail_at_scale_async(self, filename, size, callback, index,
                                     ignore_cache=False):
        """Create the thumbnail for 'filename' and return it via 'callback'.

        Creates the thumbnail for the given filename at the given size and
        then calls the given callback function with the resulting pixbuf.

        Args:
            filename: The filename to get the thumbnail for
            size: The size the returned pixbuf is scaled to
            callback: A callable of form callback(pixbuf, *args)
            args: Any additional arguments that can be passed to callback
            ignore_cache: If true, the builtin in-memory cache is bypassed and
                          the thumbnail file is loaded from disk
        """
        self._thread_pool.apply_async(self._do_get_thumbnail_at_scale,
                                      (filename, size, callback, index,
                                       ignore_cache),
                                      callback=self._do_callback)


class ThumbnailStore(object):
    """Implements freedesktop.org's Thumbnail Managing Standard."""

    KEY_URI = "Thumb::URI"
    KEY_MTIME = "Thumb::MTime"
    KEY_SIZE = "Thumb::Size"
    KEY_WIDTH = "Thumb::Image::Width"
    KEY_HEIGHT = "Thumb::Image::Height"

    def __init__(self, large=True):
        """Construct a new ThumbnailStore.

        Args:
            large: Size of thumbnails that are created. If true 256x256 else
                   128x128.
        """
        super(ThumbnailStore, self).__init__()
        import vimiv
        self.base_dir = os.path.join(GLib.get_user_cache_dir(), "thumbnails")
        self.fail_dir = os.path.join(
            self.base_dir, "fail", "vimiv-" + vimiv.__version__)
        self.thumbnail_dir = ""
        self.thumb_size = 0
        self.use_large_thumbnails(large)
        self._ensure_dirs_exist()

    def use_large_thumbnails(self, enabled=True):
        """Specify whether this thumbnail store uses large thumbnails.

        Large thumbnails have 256x256 pixels and non-large thumbnails 128x128.

        Args:
            enabled: If true large thumbnails will be used.
        """
        if enabled:
            self.thumbnail_dir = os.path.join(self.base_dir, "large")
            self.thumb_size = 256
        else:
            self.thumbnail_dir = os.path.join(self.base_dir, "normal")
            self.thumb_size = 128

    def get_thumbnail(self, filename, ignore_current=False):
        """Get the path of the thumbnail of the given filename.

        If the requested thumbnail does not yet exist, it will first be created
        before returning its path.

        Args:
            filename: The filename to get the thumbnail for.
            ignore_current: If True, ignore saved thumbnails and force a
                recreation. Needed as transforming images from within thumbnail
                mode may happen faster than in 1s.

        Return:
            The path of the thumbnail file or None if thumbnail creation failed.
        """
        # Don't create thumbnails for thumbnail cache
        if filename.startswith(self.base_dir):
            return filename

        thumbnail_filename = self._get_thumbnail_filename(filename)
        thumbnail_path = self._get_thumbnail_path(thumbnail_filename)
        if os.access(thumbnail_path, os.R_OK) \
                and self._is_current(filename, thumbnail_path) \
                and not ignore_current:
            return thumbnail_path

        fail_path = self._get_fail_path(thumbnail_filename)
        if os.path.exists(fail_path):
            # We already tried to create a thumbnail for the given file but
            # failed; don't try again.
            return None

        if self._create_thumbnail(filename, thumbnail_filename):
            return thumbnail_path

        return None

    def _ensure_dirs_exist(self):
        os.makedirs(self.thumbnail_dir, 0o700, exist_ok=True)
        os.makedirs(self.fail_dir, 0o700, exist_ok=True)

    def _is_current(self, source_file, thumbnail_path):
        source_mtime = str(self._get_source_mtime(source_file))
        thumbnail_mtime = self._get_thumbnail_mtime(thumbnail_path)
        return source_mtime == thumbnail_mtime

    def _get_thumbnail_filename(self, filename):
        uri = self._get_source_uri(filename)
        return hashlib.md5(bytes(uri, "UTF-8")).hexdigest() + ".png"

    @staticmethod
    def _get_source_uri(filename):
        return "file://" + os.path.abspath(os.path.expanduser(filename))

    def _get_thumbnail_path(self, thumbnail_filename):
        return os.path.join(self.thumbnail_dir, thumbnail_filename)

    def _get_fail_path(self, thumbnail_filename):
        return os.path.join(self.fail_dir, thumbnail_filename)

    @staticmethod
    def _get_source_mtime(src):
        return int(os.path.getmtime(src))

    def _get_thumbnail_mtime(self, thumbnail_path):
        pixbuf = Pixbuf.new_from_file(thumbnail_path)
        mtime = pixbuf.get_options()["tEXt::" + self.KEY_MTIME]

        return mtime

    def _create_thumbnail(self, source_file, thumbnail_filename):
        # Cannot access source; create neither thumbnail nor fail file
        if not os.access(source_file, os.R_OK):
            return False

        try:
            image = Pixbuf.new_from_file_at_scale(source_file, self.thumb_size,
                                                  self.thumb_size, True)
            dest_path = self._get_thumbnail_path(thumbnail_filename)
            success = True
        except GError:
            image = Pixbuf.new(GdkPixbuf.Colorspace.RGB, False, 8, 1, 1)
            dest_path = self._get_fail_path(thumbnail_filename)
            success = False

        width = 0
        height = 0
        try:
            _, width, height = GdkPixbuf.Pixbuf.get_file_info(source_file)
        except IOError:
            pass

        options = {
            "tEXt::" + self.KEY_URI: str(self._get_source_uri(source_file)),
            "tEXt::" + self.KEY_MTIME: str(self._get_source_mtime(source_file)),
            "tEXt::" + self.KEY_SIZE: str(os.path.getsize(source_file))
        }

        if width > 0 and height > 0:
            options["tEXt::" + self.KEY_WIDTH] = str(width)
            options["tEXt::" + self.KEY_HEIGHT] = str(height)

        # First create temporary file and then move it. This avoids problems
        # with concurrent access of the thumbnail cache, since "move" is an
        # atomic operation
        handle, tmp_filename = tempfile.mkstemp(dir=self.base_dir)
        os.close(handle)
        os.chmod(tmp_filename, 0o600)
        image.savev(tmp_filename, "png", list(options.keys()),
                    list(options.values()))
        os.replace(tmp_filename, dest_path)

        return success
